/* blending.c	-- Yasuhiro Yamazaki Tue Nov  6 2001
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "vector3d.h"
#include "pngtexture.h"

/* ɥξ֤򼨤 */
#define WINDOW_STATE_SHOW 0
#define WINDOW_STATE_HIDE 1

/* ž ɥʬɥåȤɤž뤫򼨤 */
#define ROTATION_COEFFICIENT 180.0 /* degree */

/* եȥ˥᡼Ф */
#define FONT_ANIMATION_MAX 3

/* ѡƥκ */
#define PARTICLE_ANIMATION_MAX 256

/* åѡƥ뤬и */
#define PARTICLE_ANIMATION_FOUNTAIN_MAX 32

/*  */
#define GRAVITY_SCALAR 1.0

/* ؿΥޥ */
#define RNDF(X) ((float)((float)rand()/(float)RAND_MAX)*(float)(X))

/* Structure: Font Animation */
typedef struct tagFontAnimation
{
  Vector3D position;
  int phase;
  double phase_timer, fade, fade_factor, fade_speed;
  double scale, scale_speed;
} FontAnimation;

/* Structure: Particle Animation */
typedef struct tagParticleAnimation
{
  Vector3D position, velocity;
  double life_timer, fade, fade_speed;
} ParticleAnimation;

/* Static ѿ */
static int window_state = WINDOW_STATE_SHOW;   /* ɥξ */
static int window_width, window_height;        /* ɥΥ */
static double ratio_x = 0.0, ratio_y = 0.0;    /* ޥɸη׻ݻ */
static float rotate_x = 0.0f, rotate_y = 0.0f; /* ɽǥβž */
static int push_button;                        /* 줿ܥݻ */

static GLuint font_tex = 0, font_tex2 = 0, particle = 0;
static FontAnimation *font_animation[ FONT_ANIMATION_MAX ];

static ParticleAnimation *particle_animation[ PARTICLE_ANIMATION_MAX ];
static int particle_max = 0;

/* Font Animation Functions */
static FontAnimation *
createFontAnimation( void )
{
  FontAnimation *ret = (FontAnimation *)malloc( sizeof( FontAnimation ) );
  setVector3DCoordinatef( &ret->position, 0.0f, 0.0f, 0.0f );
  ret->phase = 0;
  ret->phase_timer = 0.0;
  ret->fade = 0.0;
  ret->fade_factor = 3.0;
  ret->fade_speed = 1.5;
  ret->scale = 1.0;
  ret->scale_speed = 0.1;
  return ret;
}

static void
destroyFontAnimation( FontAnimation *f )
{
  free( f );
}

static void
updateFontAnimation( FontAnimation *f, const double sec )
{
  float x, y;
  switch( f->phase ){
  case 0:
    /*  */
    f->phase = 1;
    f->fade  = 0.0;
    f->scale = RNDF( 0.5f ) + 0.5f;
    x = RNDF( 2.0f )-1.0f;
    y = RNDF( 2.0f )-1.0f;
    setVector3DCoordinatef( &f->position, x, y, 0.0f );
    break;
  case 1:
    /* եɥ */
    f->fade += sec * f->fade_speed;
    if( f->fade>1.0 ){
      f->phase = 2;
      f->phase_timer = (double)RNDF( 1.0f ) + 2.5;
      f->fade = 1.0;
    }
    break;
  case 2:
    /*  */
    f->phase_timer -= sec;
    if( f->phase_timer<0.0 ){
      f->phase = 3;
    }
    break;
  case 3:
    /* եɥ */
    f->fade -= sec * f->fade_speed;
    if( f->fade<0.0 ){
      f->fade = 0.0;
      f->phase = 4;
      f->phase_timer = (double)RNDF( 1.0f ) + 4.5;
    }
    break;
  case 4:
    /*  */
    f->phase_timer -= sec;
    if( f->phase_timer<0.0 ){
      f->phase = 0;
    }
    break;
  }
  f->scale -= sec * f->scale_speed;
}

static void
putFontAnimation( FontAnimation *f )
{
  double fade;
  float scale;
  
  switch( f->phase ){
  case 0:
  case 4:
    /* ɽ */
    break;
  case 1:
  case 2:
  case 3:
    /* ɽ */
    fade = pow( f->fade, 6.5 );
    scale = (float)f->scale;

    glBindTexture( GL_TEXTURE_2D, font_tex );
    glPushMatrix();
      glTranslatef( f->position.x, f->position.y, f->position.z );
      glScalef( scale, scale, scale );
      glScalef( 2.56f, 0.38f, 0.0f );
      glColor4f( 1.0f, 1.0f, 1.0f, fade );
      glBegin( GL_TRIANGLE_STRIP );
      glTexCoord2f( 1.0f, 0.02f ); glVertex3f(  0.5f,  0.5f, 0.0f );
      glTexCoord2f( 0.0f, 0.02f ); glVertex3f( -0.5f,  0.5f, 0.0f );
      glTexCoord2f( 1.0f, 1.00f ); glVertex3f(  0.5f, -0.5f, 0.0f );
      glTexCoord2f( 0.0f, 1.00f ); glVertex3f( -0.5f, -0.5f, 0.0f );
      glEnd();

      glBindTexture( GL_TEXTURE_2D, font_tex2 );
      glScalef( pow((2.0-f->fade), f->fade_factor), 1.0f, 0.0f );
      glColor4f( 1.0f, 1.0f, 1.0f, (1.0-fade)*f->fade );
      glBegin( GL_TRIANGLE_STRIP );
      glTexCoord2f( 1.0f, 0.02f ); glVertex3f(  0.5f,  0.5f, 0.0f );
      glTexCoord2f( 0.0f, 0.02f ); glVertex3f( -0.5f,  0.5f, 0.0f );
      glTexCoord2f( 1.0f, 1.00f ); glVertex3f(  0.5f, -0.5f, 0.0f );
      glTexCoord2f( 0.0f, 1.00f ); glVertex3f( -0.5f, -0.5f, 0.0f );
      glEnd();
    glPopMatrix();
    break;
  }
}

/* ѡƥ */
static void
initializeParticleAnimation( ParticleAnimation *p )
{
  setVector3DCoordinatef( &p->position, 0.0f, -0.75f, 0.0f );
  setVector3DCoordinatef( &p->velocity, 
			  RNDF(1.0f)-0.5f,  RNDF(0.5f)+1.0f, RNDF(1.0f)-0.5f );
  p->life_timer = (double)RNDF( 1.0f ) + 1.0;
  p->fade = 1.0;
  p->fade_speed = 0.5;
}

static ParticleAnimation *
createParticleAnimation( void )
{
  ParticleAnimation *ret = (ParticleAnimation *)malloc( sizeof( ParticleAnimation ) );
  initializeParticleAnimation( ret );
  return ret;
}

static void
copyParticleAnimation( ParticleAnimation *to, ParticleAnimation *from )
{
  copyVector3D( &to->position, &from->position );
  copyVector3D( &to->velocity, &from->velocity );
  to->life_timer = from->life_timer;
  to->fade       = from->fade;
  to->fade_speed = from->fade_speed;
}

static void
destroyParticleAnimation( ParticleAnimation *p )
{
  free( p );
}

static int
updateParticleAnimation( ParticleAnimation *p, const double sec )
{
  int alive = 1;
  
  Vector3D velocity;
  copyVector3D( &velocity, &p->velocity );
  multiplyVector3D( &velocity, (float)sec );
  addVector3D( &p->position, &p->position, &velocity );
  p->velocity.y -= GRAVITY_SCALAR * sec;

  p->life_timer -= sec;
  if( p->life_timer<0.0 ){
    p->fade -= p->fade_speed * sec;
    if( p->fade<0.0 ){
      alive = 0; /* Die */
    }
  }

  return alive;
}

static void
putParticleAnimation( ParticleAnimation *p )
{
  glPushMatrix();
    glBindTexture( GL_TEXTURE_2D, particle );
    glTranslatef( p->position.x, p->position.y, p->position.z );
    glRotatef( -rotate_x, 1.0f, 0.0f, 0.0f );
    glRotatef( -rotate_y, 0.0f, 1.0f, 0.0f );
    glScalef( 0.2f, 0.2f, 0.0f );
    glColor4f( 1.0f, 1.0f, 1.0f, p->fade );
    glBegin( GL_TRIANGLE_STRIP );
    glTexCoord2f( 1.0f, 0.0f ); glVertex3f(  0.5f,  0.5f, 0.0f );
    glTexCoord2f( 0.0f, 0.0f ); glVertex3f( -0.5f,  0.5f, 0.0f );
    glTexCoord2f( 1.0f, 1.0f ); glVertex3f(  0.5f, -0.5f, 0.0f );
    glTexCoord2f( 0.0f, 1.0f ); glVertex3f( -0.5f, -0.5f, 0.0f );
    glEnd();
  glPopMatrix();
}

/* Particle Animation: Utility */
static void
updateParticleAnimations( const double sec )
{
  int i = 0;
  while( i<particle_max ){
    if( !updateParticleAnimation( particle_animation[i], sec ) ){
      copyParticleAnimation( particle_animation[i], particle_animation[ particle_max-1 ] );
      particle_max--;
      i--;
    }
    i++;
  }
}

static void
spreadParticleAnimations( const double sec )
{
  int i, amount = (int)((double)PARTICLE_ANIMATION_FOUNTAIN_MAX * sec)+1;
  if( particle_max+amount>=PARTICLE_ANIMATION_MAX ){
    amount = PARTICLE_ANIMATION_MAX - particle_max;
  }
  for( i=0; i<amount; i++ ){
    initializeParticleAnimation( particle_animation[ particle_max ] );
    particle_max++;
  }
}

/*  */
static void
initializeAll( void )
{
  int i;
  
  font_tex  = createPngTexture( "blend.png"         );
  font_tex2 = createPngTexture( "blend2.png"        );
  particle  = createPngTexture( "particlelight.png" );

  for( i=0; i<FONT_ANIMATION_MAX; i++ ){
    font_animation[i] = createFontAnimation();
  }
  for( i=0; i<PARTICLE_ANIMATION_MAX; i++ ){
    particle_animation[i] = createParticleAnimation();
  }
}

/*  */
static void
freeAll( void )
{
  int i;
  
  destroyPngTexture( font_tex  );
  destroyPngTexture( font_tex2 );
  destroyPngTexture( particle  );

  for( i=0; i<FONT_ANIMATION_MAX; i++ ){
    destroyFontAnimation( font_animation[i] );
  }
  for( i=0; i<PARTICLE_ANIMATION_MAX; i++ ){
    destroyParticleAnimation( particle_animation[i] );
  }
}

/* 褬ɬפȤʤä˸ƤФ */
static void
displayCallback( void )
{
  int i;
  
  /* ХåեΥꥢ */
  glClear( GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT );

  /* Ortho 褷 */
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  /* ֥ɤͭˤ */
  glEnable( GL_BLEND );
  glBlendFunc( GL_SRC_ALPHA, GL_ONE );
  glDepthMask( GL_FALSE );
  glEnable( GL_TEXTURE_2D );
  for( i=0; i<FONT_ANIMATION_MAX; i++ ){
    putFontAnimation( font_animation[i] );
  }
  glDisable( GL_TEXTURE_2D );
  glDepthMask( GL_TRUE );
  glDisable( GL_BLEND );

  /* Perspective ɬפ */
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  gluPerspective( 60.0, (GLdouble)window_width/(GLdouble)window_height, 
		  0.01, 100.0 );

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  /*  */
  gluLookAt( 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0 );

  glPushMatrix();
    glRotatef( rotate_y, 0.0f, 1.0f, 0.0f);
    glRotatef( rotate_x, 1.0f, 0.0f, 0.0f);

    /* ֥ɤͭˤ */
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE );
    glDepthMask( GL_FALSE );
    glEnable( GL_TEXTURE_2D );
    for( i=0; i<particle_max; i++ ){
      putParticleAnimation( particle_animation[i] );
    }
    glDisable( GL_TEXTURE_2D );
    glDepthMask( GL_TRUE );
    glDisable( GL_BLEND );

  glPopMatrix();

  /* ХåեΥå */
  glutSwapBuffers();
}

/* ɥꥵ줿˸ƤФ */
static void
reshapeCallback( int width, int height )
{
  glViewport( 0, 0, width, height );
  window_width  = width;
  window_height = height;
}

/* ɥϾƤФ */
static void
idleCallback( void )
{
  int i;
  
  if( window_state==WINDOW_STATE_SHOW ){
    /* ɥ˹Ԥ򤳤˵ */
    double sec = 0.025;
    for( i=0; i<FONT_ANIMATION_MAX; i++ ){
      updateFontAnimation( font_animation[i], sec );
    }
    updateParticleAnimations( sec );
    spreadParticleAnimations( sec );
    
    /* ɬפС򤹤 */
    glutPostRedisplay();
  }
}

/* ɥ{ɽ/줿}˸ƤФ */
static void
visibilityCallback( int state )
{
  switch( state ){
  case GLUT_NOT_VISIBLE:
    window_state = WINDOW_STATE_HIDE;
    break;
  case GLUT_VISIBLE:
    window_state = WINDOW_STATE_SHOW;
    break;
  }
}

/* ޥɸΥɥФ׻ */
static void
calculationRatioOfMouseCoordinate( const int x, const int y )
{
  int width, height;
  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );
  ratio_x = (double)y/(double)height;
  ratio_y = (double)x/(double)width;
}

/* ޥΥܥѲä˸ƤФ */
static void
mouseCallback( int button, int state, int x, int y )
{
  switch( button ){
  case GLUT_LEFT_BUTTON:
    push_button = GLUT_LEFT_BUTTON;
    break;
  case GLUT_MIDDLE_BUTTON:
    push_button = GLUT_MIDDLE_BUTTON;
    break;
  case GLUT_RIGHT_BUTTON:
    push_button = GLUT_RIGHT_BUTTON;
    break;
  }
  
  switch( state ){
  case GLUT_DOWN:
    break;
  case GLUT_UP:
    break;
  }

  calculationRatioOfMouseCoordinate( x, y );
}

/* ޥΥݥ󥿤ưä˸ƤФ */
static void
passiveMotionCallback( int x, int y )
{
  /* NOP */
}

/* ɥå桢ޥΥݥ󥿤ưä˸ƤФ */
static void
motionCallback( int x, int y )
{
  double rx, ry;
  switch( push_button ){ /* Ƥܥ˱ȿ򤵤 */
  case GLUT_LEFT_BUTTON:
    /* ʪΤβž׻ */
    rx = ratio_x;
    ry = ratio_y;
    calculationRatioOfMouseCoordinate( x, y );
    rotate_x += (float)((ratio_x-rx)*ROTATION_COEFFICIENT);
    rotate_y += (float)((ratio_y-ry)*ROTATION_COEFFICIENT);
    glutPostRedisplay();
    break;
  case GLUT_MIDDLE_BUTTON:
    break;
  case GLUT_RIGHT_BUTTON:
    break;
  }
}

/* Ϥä˸ƤФ */
static void
keyboardCallback( unsigned char key, int x, int y )
{
  switch( key ){
  /*
  ** Esc줿ץλ
  */
  case 27:
    exit( EXIT_SUCCESS );
    break;

  default:
    break;
  }
}

/* OpenGL ν */
static void
initGL( void )
{
  int width, height;

  width  = glutGet( GLUT_WINDOW_WIDTH );
  height = glutGet( GLUT_WINDOW_HEIGHT );

  reshapeCallback( width, height );

  glClearColor( 0.0, 0.0, 0.0, 1.0 );
  glEnable( GL_DEPTH_TEST );
  glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
}

/* MAIN */
int
main( int argc, char *argv[] )
{
  glutInit( &argc, argv );
  glutInitDisplayMode( GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH );

  glutCreateWindow( "Demo: BLENDing" );

  glutDisplayFunc      ( displayCallback       );
  glutKeyboardFunc     ( keyboardCallback      );
  glutReshapeFunc      ( reshapeCallback       );
  glutIdleFunc         ( idleCallback          );
  glutVisibilityFunc   ( visibilityCallback    );
  glutMouseFunc        ( mouseCallback         );
  glutPassiveMotionFunc( passiveMotionCallback );
  glutMotionFunc       ( motionCallback        );
  
  initGL(); /* OpenGL ν */
  initializeAll(); /*  */
  
  glutMainLoop(); /* ᥤ롼פγ */

  freeAll(); /*  */
  
  return EXIT_SUCCESS;
}
