/* pngtexture.c	-- Yasuhiro Yamazaki Tue Nov  6 2001
 */
#include <stdio.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glpng.h>

#include "pngtexture.h"

#define TEXTURE_SIZE_MAX 256

static int
isPowerOfTwo( const unsigned int x )
{
  if( x<2 || x&(x-1) ) return 0;
  return 1;
}

static int
round( const unsigned int x )
{
  int y = 16;
  while( x>y && y<TEXTURE_SIZE_MAX ){
    y <<= 1;
  }
  return y;
}

GLuint
createPngTexture( const char *filename )
{
  pngRawInfo png;
  GLuint texture;

  /* ƥޤ */
  glGenTextures( 1, &texture );

  /* ƥꤷޤ */
  glBindTexture( GL_TEXTURE_2D, texture );

  /* ƥΥѥ᡼ꤷޤɬĴ٤Ʋ */
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP      );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP      );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );

  /* ƥβǡɤߤޤ */
  if( pngLoadRaw( filename, &png ) ){
    int old_width = png.Width, old_height = png.Height;
    GLuint format = GL_RGB;
    int depth = 3;
    if( png.Alpha ){
      format = GL_RGBA;
      depth = 4;
    }

    if( !isPowerOfTwo( old_width ) ||
	!isPowerOfTwo( old_height ) ){
      unsigned char *make;
      png.Width = round( old_width );
      png.Height = round( old_height );
      make = (unsigned char *)malloc( png.Width * png.Height * depth );
      
      /*  OpenGL εʤ˽ޤ */
      gluScaleImage( format, old_width, old_height, GL_UNSIGNED_BYTE, png.Data,
		     png.Width, png.Height, GL_UNSIGNED_BYTE, make );

      free( png.Data );
      png.Data = make;
    }
    
    /* äǡϤޤ */
    glTexImage2D( GL_TEXTURE_2D, 0, format, 
		  png.Width, png.Height, 0, format, 
		  GL_UNSIGNED_BYTE, png.Data );

    free( png.Data );
    free( png.Palette );
  }
  
  return texture;
}

void
destroyPngTexture( const GLuint texture )
{
  glDeleteTextures( 1, &texture );
}
